/*
 *  SPDX-License-Identifier: GPL-3.0-or-later
 */

#include <QList>
#include <QString>
#include <QVariant>
#include <QMap>
#include <QByteArray>
#include <QAction>

class Canvas;
class Channel;
class ColorDepth;
class ColorManager;
class ColorModel;
class ColorProfile;
class DockWidget;
class DockWidgetFactoryBase;
class Document;
class Filter;
class InfoObject;
class Krita;
class Node;
class Notifier;
class Resource;
class Scratchpad;
class Selection;
class View;
class Extension;
class Window;

class Shape;
class GroupShape;
class PaintLayer;
class CloneLayer;
class GroupLayer;
class FilterLayer;
class FillLayer;
class FileLayer;
class VectorLayer;
